/*----- PROTECTED REGION ID(RfInp.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        RfInp.cpp
//
// description : C++ source for the RfInp class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               RfInp are implemented in this file.
//
//
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "RfInp.h"
#include "RfInpClass.h"

#include <math.h>
#include <sys/time.h>

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	RfInp.cpp

/**
 *  RfInp class description:
 *    
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//================================================================
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//  Abort         |  abort
//  Enable        |  enable
//  Disable       |  disable
//  FastRamp      |  fast_ramp
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  vlt           |  Tango::DevDouble	Scalar
//  vltSet        |  Tango::DevDouble	Scalar
//  OutputEnable  |  Tango::DevBoolean	Scalar
//================================================================

namespace RfInp_ns
{
/*----- PROTECTED REGION ID(RfInp::namespace_starting) ENABLED START -----*/
/* clang-format on */
//	static initializations
double calibrate(const std::vector<Tango::DevDouble> &calibration,
		const double &value)
{
	return calibration[0] * value * value * value + \
		calibration[1] * value * value + \
		calibration[2] * value + \
		calibration[3];
}
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	RfInp::namespace_starting

//--------------------------------------------------------
/**
 *	Method     : RfInp::RfInp()
 *	Description: Constructors for a Tango device
 *                implementing the classRfInp
 */
//--------------------------------------------------------
RfInp::RfInp(Tango::DeviceClass *cl, std::string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(RfInp::constructor_1) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::constructor_1
}
//--------------------------------------------------------
RfInp::RfInp(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(RfInp::constructor_2) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::constructor_2
}
//--------------------------------------------------------
RfInp::RfInp(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(RfInp::constructor_3) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::constructor_3
}
//--------------------------------------------------------
RfInp::~RfInp()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method     : RfInp::delete_device()
 *	Description: will be called at device destruction or at init command
 */
//--------------------------------------------------------
void RfInp::delete_device()
{
	DEBUG_STREAM << "RfInp::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::delete_device) ENABLED START -----*/
	/* clang-format on */
	//	Delete device allocated objects
	if (worker) {
		worker->stop();
		worker->join(NULL);
	}

	if (vltADC)
		delete vltADC;
	if (enableInDIO)
		delete enableInDIO;
	if (enableOutDIO)
		delete enableOutDIO;
	if (vltDAC)
		delete vltDAC;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::delete_device
	delete[] attr_vlt_read;
	delete[] attr_vltSet_read;
	delete[] attr_OutputEnable_read;
}

//--------------------------------------------------------
/**
 *	Method     : RfInp::init_device()
 *	Description: will be called at device initialization.
 */
//--------------------------------------------------------
void RfInp::init_device()
{
	DEBUG_STREAM << "RfInp::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::init_device_before) ENABLED START -----*/
	/* clang-format on */
	initError.clear();
	vltADC = NULL;
	enableInDIO = enableOutDIO = NULL;
	vltDAC = NULL;
	worker = NULL;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::init_device_before


	//	Get the device properties from database
	get_device_property();

	attr_vlt_read = new Tango::DevDouble[1];
	attr_vltSet_read = new Tango::DevDouble[1];
	attr_OutputEnable_read = new Tango::DevBoolean[1];
	//	No longer if mandatory property not set.
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(RfInp::init_device) ENABLED START -----*/
	/* clang-format on */
	//	Initialize device
	try {
		vltADC = new ADC(adcDevice);
		enableInDIO = new DIO(enableInDevice);
		enableOutDIO = new DIO(enableOutDevice);
		vltDAC = new DAC(dacDevice);

		/* Start the poller */
		worker = new RfInpWorker(this);
		worker->start();
	} catch (Tango::DevFailed &e) {
		initError = "Initialization failed: " + string(e.errors[0].desc);
	} catch (...) {
		initError = "Initialization failed: Unknown error";
	}

	if (! initError.empty()) {
		ERROR_STREAM << initError << endl;
		set_state(Tango::UNKNOWN);
		set_status(initError);
		assert(false);
	} else {
		set_state(Tango::ON);
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::init_device
}

//--------------------------------------------------------
/**
 *	Method     : RfInp::get_device_property()
 *	Description: Read database to initialize property data members.
 */
//--------------------------------------------------------
void RfInp::get_device_property()
{
	/*----- PROTECTED REGION ID(RfInp::get_device_property_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialize property data members
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("EnableInDevice"));
	dev_prop.push_back(Tango::DbDatum("EnableOutDevice"));
	dev_prop.push_back(Tango::DbDatum("MaxVltStep"));
	dev_prop.push_back(Tango::DbDatum("adcCalibration"));
	dev_prop.push_back(Tango::DbDatum("adcDevice"));
	dev_prop.push_back(Tango::DbDatum("dacCalibration"));
	dev_prop.push_back(Tango::DbDatum("dacDevice"));
	dev_prop.push_back(Tango::DbDatum("dacSetCalibration"));
	dev_prop.push_back(Tango::DbDatum("fastRampTarget"));
	dev_prop.push_back(Tango::DbDatum("fastRampTime"));
	dev_prop.push_back(Tango::DbDatum("pollingSleep"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on RfInpClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		RfInpClass	*ds_class =
			(static_cast<RfInpClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize EnableInDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  enableInDevice;
		else {
			//	Try to initialize EnableInDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  enableInDevice;
		}
		//	And try to extract EnableInDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableInDevice;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize EnableOutDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  enableOutDevice;
		else {
			//	Try to initialize EnableOutDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  enableOutDevice;
		}
		//	And try to extract EnableOutDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableOutDevice;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize MaxVltStep from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  maxVltStep;
		else {
			//	Try to initialize MaxVltStep from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  maxVltStep;
		}
		//	And try to extract MaxVltStep value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxVltStep;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize adcCalibration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  adcCalibration;
		else {
			//	Try to initialize adcCalibration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  adcCalibration;
		}
		//	And try to extract adcCalibration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  adcCalibration;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize adcDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  adcDevice;
		else {
			//	Try to initialize adcDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  adcDevice;
		}
		//	And try to extract adcDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  adcDevice;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize dacCalibration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  dacCalibration;
		else {
			//	Try to initialize dacCalibration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  dacCalibration;
		}
		//	And try to extract dacCalibration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dacCalibration;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize dacDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  dacDevice;
		else {
			//	Try to initialize dacDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  dacDevice;
		}
		//	And try to extract dacDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dacDevice;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize dacSetCalibration from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  dacSetCalibration;
		else {
			//	Try to initialize dacSetCalibration from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  dacSetCalibration;
		}
		//	And try to extract dacSetCalibration value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dacSetCalibration;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize fastRampTarget from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fastRampTarget;
		else {
			//	Try to initialize fastRampTarget from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fastRampTarget;
		}
		//	And try to extract fastRampTarget value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fastRampTarget;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize fastRampTime from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fastRampTime;
		else {
			//	Try to initialize fastRampTime from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fastRampTime;
		}
		//	And try to extract fastRampTime value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fastRampTime;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize pollingSleep from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  pollingSleep;
		else {
			//	Try to initialize pollingSleep from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  pollingSleep;
		}
		//	And try to extract pollingSleep value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  pollingSleep;

	}

	/*----- PROTECTED REGION ID(RfInp::get_device_property_after) ENABLED START -----*/
	/* clang-format on */
	//	Check device property data members init
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method     : RfInp::check_mandatory_property()
 *	Description: For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void RfInp::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << std::endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		append_status(tms.str());
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(RfInp::check_mandatory_property) ENABLED START -----*/
		/* clang-format on */
		std::cerr << tms.str() << " for " << device_name << std::endl;
		/* clang-format off */
		/*----- PROTECTED REGION END -----*/	//	RfInp::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method     : RfInp::always_executed_hook()
 *	Description: method always executed before any command is executed
 */
//--------------------------------------------------------
void RfInp::always_executed_hook()
{
	DEBUG_STREAM << "RfInp::always_executed_hook()  " << device_name << std::endl;
	if (mandatoryNotDefined)
	{
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					get_status().c_str(),
					(const char *)"RfInp::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(RfInp::always_executed_hook) ENABLED START -----*/
	/* clang-format on */
	if (! initError.empty()) {
		ERROR_STREAM << initError << endl;
		set_status(initError);
		set_state(Tango::UNKNOWN);
		assert(false);
		return;
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method     : RfInp::read_attr_hardware()
 *	Description: Hardware acquisition for attributes
 */
//--------------------------------------------------------
void RfInp::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "RfInp::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(RfInp::read_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	check_init();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method     : RfInp::write_attr_hardware()
 *	Description: Hardware writing for attributes
 */
//--------------------------------------------------------
void RfInp::write_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "RfInp::write_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(RfInp::write_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute vlt related method
 *	Description: Cavity Voltage-Direct Setting kV
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void RfInp::read_vlt(Tango::Attribute &attr)
{
	DEBUG_STREAM << "RfInp::read_vlt(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(RfInp::read_vlt) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_vlt_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::read_vlt
}
//--------------------------------------------------------
/**
 *	Write attribute vlt related method
 *	Description: Cavity Voltage-Direct Setting kV
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void RfInp::write_vlt(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "RfInp::write_vlt(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(RfInp::write_vlt) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	check_init();

	if (fabs(attr_vlt_read[0] - w_val) > maxVltStep) 
		Tango::Except::throw_exception( "",
				"Value is above MaxVltStep property",
				"RfInp::write_vlt()");
	write_vlt(w_val);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::write_vlt
}
//--------------------------------------------------------
/**
 *	Read attribute vltSet related method
 *	Description: Input Level Setting kV
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void RfInp::read_vltSet(Tango::Attribute &attr)
{
	DEBUG_STREAM << "RfInp::read_vltSet(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(RfInp::read_vltSet) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_vltSet_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::read_vltSet
}
//--------------------------------------------------------
/**
 *	Read attribute OutputEnable related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void RfInp::read_OutputEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "RfInp::read_OutputEnable(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(RfInp::read_OutputEnable) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_OutputEnable_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::read_OutputEnable
}

//--------------------------------------------------------
/**
 *	Method     : RfInp::add_dynamic_attributes()
 *	Description: Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void RfInp::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(RfInp::add_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic attributes if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command Abort related method
 *	Description: Abort Ramp
 *
 */
//--------------------------------------------------------
void RfInp::abort()
{
	DEBUG_STREAM << "RfInp::Abort()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::abort) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	check_init();

	worker->abort();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::abort
}
//--------------------------------------------------------
/**
 *	Command Enable related method
 *
 *
 */
//--------------------------------------------------------
void RfInp::enable()
{
	DEBUG_STREAM << "RfInp::Enable()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::enable) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	check_init();

	write_vlt(0);
	write_enableOut(1);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::enable
}
//--------------------------------------------------------
/**
 *	Command Disable related method
 *
 *
 */
//--------------------------------------------------------
void RfInp::disable()
{
	DEBUG_STREAM << "RfInp::Disable()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::disable) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	check_init();

	write_vlt(0);
	write_enableOut(0);

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::disable
}
//--------------------------------------------------------
/**
 *	Command FastRamp related method
 *	Description: Start Fast Ramp
 *
 */
//--------------------------------------------------------
void RfInp::fast_ramp()
{
	DEBUG_STREAM << "RfInp::FastRamp()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(RfInp::fast_ramp) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	check_init();

	if (! attr_OutputEnable_read[0])
		Tango::Except::throw_exception( "",
				"Output is disabled",
				"RfInp::fast_ramp()");

	worker->ramp();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::fast_ramp
}
//--------------------------------------------------------
/**
 *	Method     : RfInp::add_dynamic_commands()
 *	Description: Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void RfInp::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(RfInp::add_dynamic_commands) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic commands if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	RfInp::add_dynamic_commands
}

/*----- PROTECTED REGION ID(RfInp::namespace_ending) ENABLED START -----*/
/* clang-format on */
void RfInp::check_init()
{
	if (! initError.empty())
		Tango::Except::throw_exception( "",
				initError.c_str(),
				"RfInp::check_init()");
}

void RfInp::read_vlt()
{
	double value = vltADC->readRaw();
	attr_vlt_read[0] = calibrate(adcCalibration, value);
}

void RfInp::write_vlt(const double &w_val)
{
	short calibrated_value = calibrate(dacCalibration, w_val);
	vltDAC->writeRaw(calibrated_value);
}

void RfInp::read_vltSet()
{
	double value = vltDAC->readRaw();
	attr_vltSet_read[0] = calibrate(dacSetCalibration, value);
}

void RfInp::read_enableIn()
{
	attr_OutputEnable_read[0] = enableInDIO->read();
}

void RfInp::write_enableOut(const int &w_val)
{
	enableOutDIO->write(w_val);
}

void RfInpWorker::ramp()
{
	double delta;
	device->read_vlt();
	istep = device->attr_vlt_read[0];
	delta = device->fastRampTarget - istep;
	steps = device->fastRampTime * 100; // Assume 10ms
	vstep = delta / steps;

	if (delta > 0)
		action = RAMP;
}

void RfInpWorker::abort()
{
	action = POLL;
}

void* RfInpWorker::run_undetached(void*)
{
	DEBUG_STREAM << "RfInpWorker::run() - enter" << endl;
	long i, b;

	while(cont) {
		try {
			switch(action) {
				case POLL:
					device->read_vlt();
					device->read_vltSet();
					device->read_enableIn();
					device->set_state(Tango::ON);
					usleep(device->pollingSleep * 1000);
					break;
				case RAMP:
					device->set_state(Tango::MOVING);
					for (b=0, i=1; ((cont && action == RAMP) && i <= steps); ++b, ++i) {
						struct timeval start, end, diff;
						gettimeofday(&start, NULL);
						device->write_vlt(istep + vstep * i);
						switch(b) {
							case 0:
								device->read_vlt();
								break;
							case 1:
								device->read_vltSet();
								break;
							case 2:
								device->read_enableIn();
								b = -1;
								break;
						}
						gettimeofday(&end, NULL);
						timersub(&end, &start, &diff);
						if (diff.tv_sec == 0 && diff.tv_usec < 10000)
							usleep(10000 - diff.tv_usec);
					}
					action = POLL;
					break;
			}
		} catch(Tango::DevFailed &e) {
			action = POLL;
			ERROR_STREAM << string(e.errors[0].desc) << endl;
			device->set_state(Tango::UNKNOWN);
			device->set_status(string(e.errors[0].desc));
			assert(false);
		} catch(...) {
			action = POLL;
			ERROR_STREAM << "Unknown error" << endl;
			device->set_state(Tango::UNKNOWN);
			device->set_status("Unknown error");
			assert(false);
		}
	}

	DEBUG_STREAM << "RfInpWorker::run() - exit" << endl;
	
	return NULL;
}

/*----- PROTECTED REGION END -----*/	//	RfInp::namespace_ending
} //	namespace
